//////////////////////////////////////////////////////////////////////
////                                                              ////
////  SRAM with dual WISHBONE interface                           ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  SRAM used in test bench.                                    ////
////                                                              ////
////  To Do:                                                      ////
////   None                                                       ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "general.h"


module wb_sram (
	// WISHBONE interface 1
	wb1_clk_i, wb1_rst_i,
	wb1_cyc_i, wb1_adr_i, wb1_dat_i, wb1_stb_i, wb1_we_i, wb1_sel_i,
	wb1_ack_o, wb1_err_o, wb1_rty_o, wb1_dat_o, 

	// WISHBONE interface 2
	wb2_clk_i, wb2_rst_i,
	wb2_cyc_i, wb2_adr_i, wb2_dat_i, wb2_stb_i, wb2_we_i, wb2_sel_i,
	wb2_ack_o, wb2_err_o, wb2_rty_o, wb2_dat_o, 
);

//
// Parameters
//
parameter
	aw = 16,		// width of the address bus
	words = 65536;		// total number of words in RAM
// synopsys translate_off
parameter
	initfile = "../bench/mem.txt" ;
// synopsys translate_on

//
// WISHBONE interface 1
//
input			wb1_clk_i;	// clock input
input			wb1_rst_i;	// reset input
input			wb1_cyc_i;	// cycle valid input
input	[aw-1:0]	wb1_adr_i;	// address bus inputs
input	[31:0]		wb1_dat_i;	// input data bus
input			wb1_stb_i;	// strobe input
input			wb1_we_i;	// indicates write transfer
input	[3:0]		wb1_sel_i;	// byte select inputs
output			wb1_ack_o;	// normal termination
output			wb1_err_o;	// termination w/ error
output			wb1_rty_o;	// termination w/ retry
output	[31:0]		wb1_dat_o;	// output data bus

//
// WISHBONE interface 2
//
input			wb2_clk_i;	// clock input
input			wb2_rst_i;	// reset input
input			wb2_cyc_i;	// cycle valid input
input	[aw-1:0]	wb2_adr_i;	// address bus inputs
input	[31:0]		wb2_dat_i;	// input data bus
input			wb2_stb_i;	// strobe input
input			wb2_we_i;	// indicates write transfer
input	[3:0]		wb2_sel_i;	// byte select inputs
output			wb2_ack_o;	// normal termination
output			wb2_err_o;	// termination w/ error
output			wb2_rty_o;	// termination w/ retry
output	[31:0]		wb2_dat_o;	// output data bus

//
// Registers
//
reg	[31:0] 		wb1_dat_o;	// output data bus
reg			wb1_read_ok;	// Internal read ACK
reg	[31:0]		wb1_tmp;
reg	[31:0] 		wb2_dat_o;	// output data bus
reg			wb2_read_ok;	// Internal read ACK
reg	[31:0]		wb2_tmp;
reg	[31:0]		mem [words-1:0];// memory core
integer			i;

//
// Transfer termination assignments
//
assign wb1_ack_o = wb1_stb_i & wb1_we_i | wb1_read_ok;
assign wb1_err_o = 1'b0;
assign wb1_rty_o = 1'b0;
assign wb2_ack_o = wb2_stb_i & wb2_we_i | wb2_read_ok;
assign wb2_err_o = 1'b0;
assign wb2_rty_o = 1'b0;

//
// initialize ram if required
//
//synopsys translate_off
initial begin
	wb1_read_ok = 0;
	wb1_dat_o = 0;
	wb2_read_ok = 0;
	wb2_dat_o = 0;
	for(i = 0; i < words; i = i + 1)
		mem[i] = 0;
	if (initfile != "none") begin
		$write(" **  Initializing RAM (%m) with (%s)\n", initfile ) ;
		$readmemh(initfile, mem, 0, 5000) ;
	end
end
//synopsys translate_on

//
// WB1 write logic
//
always @(posedge wb1_clk_i)
	if (wb1_stb_i && wb1_we_i) begin
		wb1_tmp = mem[wb1_adr_i[aw-1:2]];
		if (wb1_sel_i[3])
			wb1_tmp[31:24] = wb1_dat_i[31:24];
		if (wb1_sel_i[2])
			wb1_tmp[23:16] = wb1_dat_i[23:16];
		if (wb1_sel_i[1])
			wb1_tmp[15:8] = wb1_dat_i[15:8];
		if (wb1_sel_i[0])
			wb1_tmp[7:0] = wb1_dat_i[7:0];
		mem[wb1_adr_i[aw-1:2]] = wb1_tmp;
		$display("write: sel %h (new mem %h)", wb1_sel_i, mem[wb1_adr_i[aw-1:2]]);
	end

//
// WB2 write logic
//
always @(posedge wb2_clk_i)
	if (wb2_stb_i && wb2_we_i) begin
		wb2_tmp = mem[wb2_adr_i[aw-1:2]];
		if (wb2_sel_i[3])
			wb2_tmp[31:24] = wb2_dat_i[31:24];
		if (wb2_sel_i[2])
			wb2_tmp[23:16] = wb2_dat_i[23:16];
		if (wb2_sel_i[1])
			wb2_tmp[15:8] = wb2_dat_i[15:8];
		if (wb2_sel_i[0])
			wb2_tmp[7:0] = wb2_dat_i[7:0];
		mem[wb2_adr_i[aw-1:2]] = wb2_tmp;
		$display("write: sel %h (new mem %h)", wb2_sel_i, mem[wb2_adr_i[aw-1:2]]);
	end

//
// WB1 read logic
//
always @(posedge wb1_clk_i)
	if (wb1_stb_i && !wb1_we_i) begin
		wb1_dat_o = mem[wb1_adr_i[aw-1:2]];
		wb1_read_ok = 1'b1;
	end
	else
		wb1_read_ok = 1'b0;

//
// WB2 read logic
//
always @(posedge wb2_clk_i)
	if (wb2_stb_i && !wb2_we_i) begin
		wb2_dat_o = mem[wb2_adr_i[aw-1:2]];
		wb2_read_ok = 1'b1;
	end
	else
		wb2_read_ok = 1'b0;

endmodule // wb_sram
