//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200's Data TLB                                           ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  Data TLB                                                    ////
////                                                              ////
////  To Do:                                                      ////
////   - make it smaller and faster                               ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "general.h"

`define VPN_TAG		31:20
`define VPN_INDEX	19:13
`define VPN		31:13
`define PPN		31:13
`define PTE_VPN		31:20
`define PTE_PPN		31:13
`define PTE_PPN1	31:27
`define PTE_PPN2	26:13
`define PTE_FLAGS	9:4
`define TLB_A		0
`define TLB_D		1
`define TLB_URE		2
`define TLB_UWE		3
`define TLB_SRE		4
`define TLB_SWE		5
`define TLB_PPN2	19:6
`define TLB_PPN1	4:0
`define TLB_VPTAG	16:5
`define MOD_INDEX	6:0

// DTLB unit
module dtlb(clk, rst, tlbtr_we, tlbmr_we, do_transl, tlb_en, access_type, vpn, saved_vpn,
	mod_datain, mod_addr,
	ppn, tlb_miss, pgfault
);

input clk;
input rst;
input tlbtr_we;
input tlbmr_we;
input do_transl;
input tlb_en;
input [`ACCESS_WIDTH-1:0] access_type;
input [`VPN] vpn;
input [`VPN] saved_vpn;
input [31:0] mod_datain;
input [31:0] mod_addr;

output [`PPN] ppn;
output tlb_miss;
output pgfault;

wire [16:0] to_ram17;
wire [19:0] to_ram20;
wire [16:0] wen17;
wire [19:0] wen20;
wire [16:0] from_ram17;
wire [19:0] from_ram20;
wire cen17;
wire cen20;
wire [6:0] addr17;
wire [6:0] addr20;

reg transl_done;

// mod_datain[31:13] = PPN
// mod_datain[9] = SWE
// mod_datain[8] = SRE
// mod_datain[7] = UWE
// mod_datain[6] = URE
// mod_datain[5] = dirty
// mod_datain[4] = access
assign to_ram17 = { mod_datain[`PTE_VPN], mod_datain[`PTE_PPN1]};
assign to_ram20 = { mod_datain[`PTE_PPN2], mod_datain[`PTE_FLAGS]};

// TLBMR and TLBTR are not in seperate memory blocks so we need
// properly write into memory blocks
assign wen17 = {{13{~tlbmr_we}}, {5{~tlbtr_we}}};
assign wen20 = {20{~tlbtr_we}};

// CEs for memories: enable either when doing translations or when
// writing into TLB memories
assign cen17 = ~(do_transl | tlbmr_we | tlbtr_we);
assign cen20 = ~(do_transl | tlbtr_we);

// TLB indexing: when TLB is disabled, it is indexed by mod_addr.
// When TLB is enabled, it is indexed by VPN_INDEX.
assign addr17 = tlb_en ? vpn[`VPN_INDEX] : mod_addr[`MOD_INDEX];
assign addr20 = tlb_en ? vpn[`VPN_INDEX] : mod_addr[`MOD_INDEX];

// If TLB disabled then PPN = SAVED_VPN
assign ppn = (tlb_en == 1'b1) ? {from_ram17[`TLB_PPN1], from_ram20[`TLB_PPN2]} : saved_vpn;

// Comparison of stored TLB tag with current VADDR's page index
// Lower 7 bits of VADDR's page index are ignored since we index
// TLB entries with them so they always match
assign tlb_miss = (transl_done && (from_ram17[`TLB_VPTAG] != saved_vpn[`VPN_TAG])) ? 1'b1 : 1'b0;

// Comparison of access permission bits
assign pgfault = (transl_done &&
			~((from_ram20[`TLB_URE] && access_type == `ACCESS_USER_READ) ||
			 (from_ram20[`TLB_UWE] && access_type == `ACCESS_USER_WRITE) ||
			 (from_ram20[`TLB_SRE] && access_type == `ACCESS_SUPV_READ) ||
			 (from_ram20[`TLB_SWE] && access_type == `ACCESS_SUPV_WRITE)));

// Asserted when translation is done (which is always one cycle after do_transl
// is asserted)
always @(posedge rst or posedge clk)
	if (rst)
		transl_done <= #1 1'b0;
	else
		transl_done <= #1 do_transl & tlb_en;

art_hdsp_128x17 dtlb0 (
	.q(from_ram17),
	.clk(clk),
	.cen(cen17),
	.wen(wen17),
	.a(addr17),
	.d(to_ram17)
);

art_hdsp_128x20 dtlb1 (
	.q(from_ram20),
	.clk(clk),
	.cen(cen20),
	.wen(wen20),
	.a(addr20),
	.d(to_ram20)
);

endmodule
