//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200's Instruction Cache Top Level                        ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  Instruction cache top level instantiating all IC blocks.    ////
////                                                              ////
////  To Do:                                                      ////
////   - make it smaller and faster                               ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "general.h"

// Insn cache

module ic(clk, rst, clkdiv_by_2, icbiu_rdy, icfetch_addr, icfetch_op, icbiu_datain,
	tp4w, tpdw,
	icfetch_dataout, icbiu_addr, icbiu_read, icfetch_stall, mod_dataout
);

parameter dw = `OPERAND_WIDTH;

input clk;
input rst;
input clkdiv_by_2;
input icbiu_rdy;
input [31:0] icfetch_addr;
input [`FETCHOP_WIDTH-1:0] icfetch_op;
input [dw-1:0] icbiu_datain;
input [dw-1:0] mod_dataout;

input [`TP4W_WIDTH-1:0] tp4w;
input [31:0] tpdw;

output [dw-1:0] icfetch_dataout;
output [31:0] icbiu_addr;
output icbiu_read;
output icfetch_stall;

wire [18:0] tag;
wire [dw-1:0] to_icram;
wire [dw-1:0] from_icram;
wire [31:0] saved_addr;
wire refill;
wire [3:0] icram_we;
wire ictag_we;
wire [31:0] ic_addr;
wire refill_first;
wire refill_prepare;
wire refill_start;
wire refill_rest;

reg [1:0] valid_div;
reg hit;

wire queue;
wire cntrbusy;
wire icbiu_valid;

assign icbiu_addr = ic_addr;
assign ictag_we = refill;

// assign ic_stall = icbiu_read | refill_start;  // can't remember if it works
assign queue = (refill && icfetch_op && !refill_first & !refill_rest) ? 1'b1 : 1'b0;

// assign icfetch_stall = refill_start | (refill_first & ~icbiu_valid)| refill_rest | queue;  // kind of working
assign icfetch_stall = refill | ~hit;
//assign icfetch_stall = refill;

// Select between claddr generated by IC FSM and addr[3:2] generated by IFETCH
assign ic_addr = (refill == 1'b1) ? saved_addr : icfetch_addr;

// Input data generated by BIU
assign to_icram = icbiu_datain;

// Select between data generated by ICRAM or passed by BIU
assign icfetch_dataout = (refill_first == 1'b1) ? icbiu_datain : from_icram;

// Tag comparison
always @(tag or saved_addr) begin
        if (tag == saved_addr[31:13])
                hit <= #1 1'b1;
        else
                hit <= #1 1'b0;
end

// Valid_div counts RISC clock cycles by modulo 4
always @(posedge clk or posedge rst)
	if (rst)
		valid_div <= #1 2'b0;
	else
		valid_div <= #1 valid_div + 'd1;

// icbiu_valid is one RISC clock cycle long icbiu_rdy.
// icbiu_rdy is two/four RISC clock cycles long because memory
// controller works at 1/2 or 1/4 of RISC clock freq (at 1/2 if
// clkdiv_by_2 is asserted).
assign icbiu_valid = icbiu_rdy & (valid_div[1] | clkdiv_by_2) & valid_div[0];

// Generate refill_start that signals to frz_logic a cache linefill is about to begin
assign refill_start = (refill_prepare & ~hit) ? 1'b1 : 1'b0;

itlb itlb(
        .clk(clk),
        .rst(rst),
        .we(1'b0),
        .tlb_en(1'b1),
        .datain(mod_dataout),
        .vaddr(saved_addr),
        .paddr(),
        .cache_en(),
        .tlb_miss(),
	.tp4w(tp4w),
	.tpdw(tpdw)
);

ic_fsm ic_fsm(
	.clk(clk),
	.rst(rst),
	.fetch_op(icfetch_op),
	.miss(~hit),
	.biudata_valid(icbiu_valid),
	.start_addr(icfetch_addr),
	.saved_addr(saved_addr),
	.refill(refill),
	.refill_first(refill_first),
	.refill_prepare(refill_prepare),
	.icram_we(icram_we),
	.biu_read(icbiu_read),
	.refill_rest(refill_rest),
	.cntrbusy(cntrbusy)
);

ic_ram ic_ram(
	.clk(clk),
	.addr(ic_addr[12:2]),
	.we(icram_we),
	.datain(to_icram),
	.dataout(from_icram),
	.tp4w(tp4w),
	.tpdw(tpdw)
);

ic_tag ic_tag(
	.clk(clk),
	.addr(ic_addr[12:4]),
	.we(ictag_we),
	.datain(ic_addr[31:13]),
	.dataout(tag),
	.tp4w(tp4w),
	.tpdw(tpdw)
);

endmodule
