//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200 Top Level                                            ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  OR1200 Top Level                                            ////
////                                                              ////
////  To Do:                                                      ////
////   - make it smaller and faster                               ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "general.h"

module or1200(
	clk, rst, irq, clkdiv_by_2,

	// Instruction WISHBONE INTERFACE
	iwb_clk_i, iwb_rst_i, iwb_ack_i, iwb_err_i, iwb_rty_i, iwb_dat_i,
	iwb_cyc_o, iwb_adr_o, iwb_stb_o, iwb_we_o, iwb_sel_o, iwb_dat_o,

	// Data WISHBONE INTERFACE
	dwb_clk_i, dwb_rst_i, dwb_ack_i, dwb_err_i, dwb_rty_i, dwb_dat_i,
	dwb_cyc_o, dwb_adr_o, dwb_stb_o, dwb_we_o, dwb_sel_o, dwb_dat_o,

	tp_dir_in, tp_sel, tp_in, tp_out,
	mod_addr, mod_from_risc, mod_to_risc
);

parameter dw = `OPERAND_WIDTH;
parameter aw = `OPERAND_WIDTH;

//
// Instruction WISHBONE interface
//
input		iwb_clk_i;	// clock input
input		iwb_rst_i;	// reset input
input		iwb_ack_i;	// normal termination
input		iwb_err_i;	// termination w/ error
input		iwb_rty_i;	// termination w/ retry
input [dw-1:0]	iwb_dat_i;	// input data bus
output		iwb_cyc_o;	// cycle valid output
output [aw-1:0]	iwb_adr_o;	// address bus outputs
output		iwb_stb_o;	// strobe output
output		iwb_we_o;	// indicates write transfer
output [3:0]	iwb_sel_o;	// byte select outputs
output [dw-1:0]	iwb_dat_o;	// output data bus

//
// Data WISHBONE interface
//
input		dwb_clk_i;	// clock input
input		dwb_rst_i;	// reset input
input		dwb_ack_i;	// normal termination
input		dwb_err_i;	// termination w/ error
input		dwb_rty_i;	// termination w/ retry
input [dw-1:0]	dwb_dat_i;	// input data bus
output		dwb_cyc_o;	// cycle valid output
output [aw-1:0]	dwb_adr_o;	// address bus outputs
output		dwb_stb_o;	// strobe output
output		dwb_we_o;	// indicates write transfer
output [3:0]	dwb_sel_o;	// byte select outputs
output [dw-1:0]	dwb_dat_o;	// output data bus

//
// System
//
input clk;
input rst;
input clkdiv_by_2;

// Incoming interrupts
input irq;

// These connect DC to BIU
wire dcbiu_rdy;
wire [31:0] dcbiu_from_biu;
wire [31:0] dcbiu_to_biu;
wire [31:0] dcbiu_addr;
wire dcbiu_read;
wire dcbiu_write;

// These connect IC to BIU
wire icbiu_rdy;
wire [31:0] icbiu_from_biu;
wire [31:0] icbiu_addr;
wire icbiu_read;

// These connect RISC's module interface to peripherals
output [31:0] mod_addr;
output [31:0] mod_from_risc;
input [31:0] mod_to_risc;

// These connect DC and CPU's LSU
wire dclsu_stall;
wire [31:0] dclsu_addr;
wire [31:0] dclsu_from_dc;
wire [31:0] dclsu_to_dc;
wire [`LSUOP_WIDTH-1:0] dclsu_lsuop;

// These connect IC and CPU's ifetch
wire icfetch_stall;
wire [31:0] icfetch_addr;
wire [31:0] icfetch_dataout;

// These connect INTC and CPU's EXCEPT
wire int_high;
wire int_low;

// These connect CPU and external Trace port
input tp_dir_in;
input [1:0] tp_sel;
input [31:0] tp_in;
output [31:0] tp_out;

// These conenct Trace port and caches/MMUs
wire [`TP2W_WIDTH-1:0] tp2w;
wire [`TP3W_WIDTH-1:0] tp3w;
wire [`TP4W_WIDTH-1:0] tp4w;
wire [31:0] tpdw;

//
// Instantiation of Instruction WISHBONE BIU
//
wb_biu iwb_biu(
	// WISHBONE interface
	.wb_clk_i(iwb_clk_i),
	.wb_rst_i(iwb_rst_i),
	.wb_ack_i(iwb_ack_i),
	.wb_err_i(iwb_err_i),
	.wb_rty_i(iwb_rty_i),
	.wb_dat_i(iwb_dat_i),
	.wb_cyc_o(iwb_cyc_o),
	.wb_adr_o(iwb_adr_o),
	.wb_stb_o(iwb_stb_o),
	.wb_we_o(iwb_we_o),
	.wb_sel_o(iwv_sel_o),
	.wb_dat_o(iwb_dat_o),

	// Internal RISC bus
	.biu_to_biu(32'b0),
	.biu_addr(icbiu_addr),
	.biu_read(icbiu_read),
	.biu_write(1'b0),
	.biu_rdy(icbiu_rdy),
	.biu_from_biu(icbiu_from_biu)
);

//
// Instantiation of Data WISHBONE BIU
//
wb_biu dwb_biu(
	// WISHBONE interface
	.wb_clk_i(dwb_clk_i),
	.wb_rst_i(dwb_rst_i),
	.wb_ack_i(dwb_ack_i),
	.wb_err_i(dwb_err_i),
	.wb_rty_i(dwb_rty_i),
	.wb_dat_i(dwb_dat_i),
	.wb_cyc_o(dwb_cyc_o),
	.wb_adr_o(dwb_adr_o),
	.wb_stb_o(dwb_stb_o),
	.wb_we_o(dwb_we_o),
	.wb_sel_o(dwv_sel_o),
	.wb_dat_o(dwb_dat_o),

	// Internal RISC bus
	.biu_to_biu(dcbiu_to_biu),
	.biu_addr(dcbiu_addr),
	.biu_read(dcbiu_read),
	.biu_write(dcbiu_write),
	.biu_rdy(dcbiu_rdy),
	.biu_from_biu(dcbiu_from_biu)
);

ic ic(
	.clk(clk),
	.rst(rst),
	.clkdiv_by_2(clkdiv_by_2),

	// These connect IC to CPU's ifetch
	.icfetch_addr(icfetch_addr),
	.icfetch_op(`FETCHOP_LW),
	.icfetch_dataout(icfetch_dataout),
	.icfetch_stall(icfetch_stall),

	// These connect IC to BIU
	.icbiu_rdy(icbiu_rdy),
	.icbiu_datain(icbiu_from_biu),
	.icbiu_addr(icbiu_addr),
	.icbiu_read(icbiu_read),

	// These connect IC to SPRS
	.mod_dataout(mod_from_risc),

	// Trace port
	.tp4w(tp4w),
	.tpdw(tpdw)
);

cpu cpu(
	.clk(clk),
	.rst(rst),

	// These connect IC and IFETCHER inside CPU
	.ic_insn(icfetch_dataout),
	.ic_pcaddr(icfetch_addr),
	.ic_stall(icfetch_stall),

	// These connect CPU to external Trace port
	.tp_dir_in(tp_dir_in),
	.tp_sel(tp_sel),
	.tp_in(tp_in),
	.tp_out(tp_out),

	// These connect DC and CPU's LSU
	.dclsu_stall(dclsu_stall),
	.dclsu_addr(dclsu_addr),
	.dclsu_datain(dclsu_from_dc),
	.dclsu_dataout(dclsu_to_dc),
	.dclsu_lsuop(dclsu_lsuop),

	// These connect INTC and CPU's EXCEPT
	.int_high(int_high),
	.int_low(int_low),

	// These connect module interface to peripherals
	.mod_addr(mod_addr),
	.mod_dataout(mod_from_risc),
	.mod_datain(mod_to_risc),

	// These connect trace port to caches and MMUs
	.tp2w(tp2w),
	.tp3w(tp3w),
	.tp4w(tp4w),
	.tpdw(tpdw)
);

dc dc(
	.clk(clk),
	.rst(rst),
	.clkdiv_by_2(clkdiv_by_2),

	// These connect DC to CPU's LSU
	.dclsu_addr(dclsu_addr),
	.dclsu_lsuop(dclsu_lsuop),
	.dclsu_datain(dclsu_to_dc),
	.dclsu_dataout(dclsu_from_dc),
	.dclsu_stall(dclsu_stall),

	// These connect DC to BIU
	.dcbiu_rdy(dcbiu_rdy),
	.dcbiu_datain(dcbiu_from_biu),
	.dcbiu_dataout(dcbiu_to_biu),
	.dcbiu_addr(dcbiu_addr),
	.dcbiu_read(dcbiu_read),
	.dcbiu_write(dcbiu_write),

	// Trace port
	.tp2w(tp2w),
	.tp3w(tp3w),
	.tpdw(tpdw)
);


//always begin
//	int_low <= #1 1'b0;
//	#1500;
//	int_low <= #1 1'b1;
//	#100 int_low <= #1 1'b0;
//
//	#1000 finish_simulation;
//end


intc intc(
	.clk(clk),
	.rst(rst),
	.irq(irq),
	.int_low(int_low),
	.int_high(int_high),
        .cs(1'b0),
	.addr(2'b0),
	.datain()
);

endmodule

