//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200's interface to SPRs                                  ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  Decoding of SPR addresses and access to SPRs                ////
////                                                              ////
////  To Do:                                                      ////
////   - make it smaller and faster                               ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "alu_defs.h"
`include "general.h"


module sprs(clk, rst, a, b, alu_op, flag, dtlb_we, to_wbmux, mod_addr, mod_datain, mod_dataout,
		epcr_we, eear_we, esr_we, epcr, eear, esr,
		sr, except_start, except_started, branch_op);

parameter width = `OPERAND_WIDTH;

input clk;
input rst;
input flag;
input [width-1:0] a;
input [width-1:0] b;
input [`ALUOP_WIDTH-1:0] alu_op;
input [`BRANCHOP_WIDTH-1:0] branch_op;
output dtlb_we;
output [width-1:0] to_wbmux;

output [31:0] mod_addr;
output [31:0] mod_dataout;
input [31:0] mod_datain;

output epcr_we;
output eear_we;
output esr_we;
input [31:0] epcr;
input [31:0] eear;
input [`SR_WIDTH-1:0] esr;

output [`SR_WIDTH-1:0] sr;
input except_start;
input except_started;

reg [31:0] mod_addr;
reg [31:0] mod_dataout;
reg [`SR_WIDTH-1:0] sr;

reg write_spr;
reg read_spr;
reg [width-1:0] to_wbmux;
wire dtlb_we;
wire itlb_we;
wire mod_addr_we;
wire mod_dataout_we;
wire sr_we;
wire sel_sys;
wire sr_sel;
wire epcr_sel;
wire eear_sel;
wire esr_sel;
wire [31:0] sys_data;
wire [`SR_WIDTH-1:0] to_sr;

assign to_sr = (branch_op == `BRANCHOP_RFE) ? esr : b[`SR_WIDTH-1:0];
 
assign dtlb_we = (write_spr && (a[`SPR_GROUP_BITS] == `SPR_GROUP_DTLB));
assign itlb_we = (write_spr && (a[`SPR_GROUP_BITS] == `SPR_GROUP_ITLB));
assign mod_dataout_we = (write_spr && (a[`SPR_GROUP_BITS] == `SPR_GROUP_MODD));
assign mod_addr_we = (write_spr && (a[`SPR_GROUP_BITS] == `SPR_GROUP_MODA));
assign sel_sys = (a[`SPR_GROUP_BITS] == `SPR_GROUP_SYS);

assign sr_sel = (sel_sys && (a[6:0] == `SPR_SR));
assign epcr_sel = (sel_sys && (a[6:0] == `SPR_EPCR));
assign eear_sel = (sel_sys && (a[6:0] == `SPR_EEAR));
assign esr_sel = (sel_sys && (a[6:0] == `SPR_ESR));

assign sr_we = (write_spr && sr_sel) | (branch_op == `BRANCHOP_RFE);
assign epcr_we = (write_spr && epcr_sel);
assign eear_we = (write_spr && eear_sel);
assign esr_we = (write_spr && esr_sel);

assign sys_data = (sr & {32{read_spr & sr_sel}}) |
		  (epcr & {32{read_spr & epcr_sel}}) |
		  (eear & {32{read_spr & eear_sel}}) |
		  (esr & {32{read_spr & esr_sel}});

// Module Dataout register
always @(posedge clk or posedge rst)
	if (rst)
		mod_dataout <= #1 32'b0;
	else if (mod_dataout_we) begin
		$display(" INFO: writing into mod_data register: %h", b);
		mod_dataout <= #1 b;
	end

// Module Addr register
always @(posedge clk or posedge rst)
	if (rst)
		mod_addr <= #1 32'b0;
	else if (mod_addr_we) begin
		$display(" INFO: writing into mod_addr register: %h", b);
		mod_addr <= #1 b;
	end

// Supervision register
always @(posedge clk or posedge rst)
	if (rst)
		sr <= #1 `SR_WIDTH'b011;
	else if (except_started) begin
		$display(" INFO: exception started. SR[SUPV] set and SR[EXR] cleared.");
		sr[`SR_SUPV] <= #1 1'b1;
		sr[`SR_EXR] <= #1 1'b0;
	end
	else if (sr_we) begin
		$display(" INFO: writing into SR register: %h", b);
		sr <= #1 to_sr;
	end

always @(alu_op or a or b or flag or sys_data or mod_datain) begin
	case (alu_op)	// synopsys full_case parallel_case
		`ALUOP_MTSR : begin
			$display("%t: SPRS: mtspr (%h) <- %h", $time, a, b);
			write_spr <= #1 1'b1;
			read_spr <= #1 1'b0;
		end
		`ALUOP_MFSR : begin
			casex (a[`SPR_GROUP_BITS])
                                5'b11111:
					to_wbmux <= #1 32'b0;
                                `SPR_GROUP_MODD:
					to_wbmux <= #1 mod_datain;
                                `SPR_GROUP_SYS:
					to_wbmux <= #1 sys_data;
                                5'b00001:
					to_wbmux <= #1 {31'b0, flag};
				default:
					to_wbmux <= #1 32'b0;
			endcase
			write_spr <= #1 1'b0;
			read_spr <= #1 1'b1;
		end
		default : begin
			write_spr <= #1 1'b0;
			read_spr <= #1 1'b0;
		end
	endcase
end


endmodule
