//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200's traceport                                          ////
////                                                              ////
////  This file is part of the OpenRISC 1200 project              ////
////  http://www.opencores.org/cores/or1k/                        ////
////                                                              ////
////  Description                                                 ////
////  Debug interface to the RIS                                  ////
////                                                              ////
////  To Do:                                                      ////
////   - transform it into OR1K compliant debug unit              ////
////                                                              ////
////  Author(s):                                                  ////
////      - Damjan Lampret, lampret@opencores.org                 ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
////                                                              ////
//// Copyright (C) 2000 Authors and OPENCORES.ORG                 ////
////                                                              ////
//// This source file may be used and distributed without         ////
//// restriction provided that this copyright statement is not    ////
//// removed from the file and that any derivative work contains  ////
//// the original copyright notice and the associated disclaimer. ////
////                                                              ////
//// This source file is free software; you can redistribute it   ////
//// and/or modify it under the terms of the GNU Lesser General   ////
//// Public License as published by the Free Software Foundation; ////
//// either version 2.1 of the License, or (at your option) any   ////
//// later version.                                               ////
////                                                              ////
//// This source is distributed in the hope that it will be       ////
//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
//// PURPOSE.  See the GNU Lesser General Public License for more ////
//// details.                                                     ////
////                                                              ////
//// You should have received a copy of the GNU Lesser General    ////
//// Public License along with this source; if not, download it   ////
//// from http://www.opencores.org/lgpl.shtml                     ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
//
// CVS Revision History
//
// $Log$
//

`include "general.h"

module traceport(
		tp_dir_in, tp_sel, tp_in, tp_out,	// Trace port connection to outside world
		rst,
		rfa_tqa, rfb_tqa, rfa_tmuxed, rfb_tmuxed,
		wb_pc, wb_insn, wb_forw,		// Trace sources from RISC core
		wr_insn, insn,
		tp1w, tp2w, tp3w, tp4w, tpdw		// To RISC core
);

// From outside world
input tp_dir_in;
input [1:0] tp_sel;
input [31:0] tp_in;

// To outside world
output [31:0] tp_out;

// From RISC core
input rst;
input [31:0] rfa_tqa;
input [31:0] rfb_tqa;
input [`TP1R_WIDTH-1:0] rfa_tmuxed;
input [`TP1R_WIDTH-1:0] rfb_tmuxed;
input [31:0] wb_pc;
input [31:0] wb_insn;
input [31:0] wb_forw;

// To RISC core
output wr_insn;
output [31:0] insn;
output [`TP1W_WIDTH-1:0] tp1w;
output [`TP2W_WIDTH-1:0] tp2w;
output [`TP3W_WIDTH-1:0] tp3w;
output [`TP4W_WIDTH-1:0] tp4w;
output [31:0] tpdw;

// Internal wires
wire wr_tp1;
wire wr_tp2;
wire wr_tp3;
wire wr_tp4;
wire wr_tpd;
wire [31:0] muxed;

// Internal registers
reg [`TP1W_WIDTH-1:0] tp1w;
reg [`TP2W_WIDTH-1:0] tp2w;
reg [`TP3W_WIDTH-1:0] tp3w;
reg [`TP4W_WIDTH-1:0] tp4w;
reg [31:0] tpdw;

// Output port.
assign tp_out = (tp_sel == 2'd0) ? wb_pc :
		(tp_sel == 2'd1) ? wb_insn :
		(tp_sel == 2'd2) ? wb_forw : muxed;

// Strobe signals when 'trace' port is input.
// Right now we can set just fetched insn.
assign wr_insn = (tp_dir_in && (tp_sel == 2'd0));

// Trace port control register 1 enable
assign wr_tp1 = (tp_dir_in && (tp_sel == 2'd1));

// Trace port control register 2 enable
assign wr_tp2 = (tp_dir_in && (tp_sel == 2'd2) && (tp1w[`TP1W_MUXSEL] == 3'd2));

// Trace port control register 3 enable
assign wr_tp3 = (tp_dir_in && (tp_sel == 2'd2) && (tp1w[`TP1W_MUXSEL] == 3'd3));

// Trace port control register 4 enable
assign wr_tp4 = (tp_dir_in && (tp_sel == 2'd2) && (tp1w[`TP1W_MUXSEL] == 3'd4));

// Trace port data register enable
assign wr_tpd = (tp_dir_in && (tp_sel == 2'd2) && (tp1w[`TP1W_MUXSEL] == 3'd5));

// Set all outputs that go to RISC core
assign insn = tp_in;

// Output muxing
assign muxed = (tp1w[`TP1W_MUXSEL] == 3'd0) ? rfa_tqa :
		(tp1w[`TP1W_MUXSEL] == 3'd1) ? rfb_tqa :
		(tp1w[`TP1W_MUXSEL] == 3'd2) ? {{32-`TP1R_WIDTH{1'b0}}, rfa_tmuxed}:
		(tp1w[`TP1W_MUXSEL] == 3'd3) ? {{32-`TP1R_WIDTH{1'b0}}, rfb_tmuxed}:
		32'b0;

// Trace port control register 1
always @(posedge wr_tp1 or posedge rst)
	if (rst) begin
		tp1w <= `TP1W_WIDTH'b0;
		tp1w[`TP1W_RF_TCENA] <= 1'b1;
		tp1w[`TP1W_RF_TCENB] <= 1'b1;
	end else
		tp1w <= #1 tp_in[`TP1W_WIDTH-1:0];

// Trace port control register 2
always @(posedge wr_tp2 or posedge rst)
	if (rst) begin
		tp2w <= `TP2W_WIDTH'b0;
		tp2w[`TP2W_DCR_TCEN] <= 1'b1;
		tp2w[`TP2W_DCT_TCEN] <= 1'b1;
	end else
		tp2w <= #1 tp_in[`TP2W_WIDTH-1:0];

// Trace port control register 3
always @(posedge wr_tp3 or posedge rst)
	if (rst) begin
		tp3w <= `TP3W_WIDTH'b0;
		tp3w[`TP3W_DMT_TCEN] <= 1'b1;
	end else
		tp3w <= #1 tp_in[`TP3W_WIDTH-1:0];

// Trace port control register 4
always @(posedge wr_tp4 or posedge rst)
	if (rst) begin
		tp4w <= `TP4W_WIDTH'b0;
		tp4w[`TP4W_IMT_TCEN] <= 1'b1;
		tp4w[`TP4W_ICT_TCEN] <= 1'b1;
	end else
		tp4w <= #1 tp_in[`TP4W_WIDTH-1:0];

// Trace port data register
always @(posedge wr_tpd or posedge rst)
	if (rst)
		tpdw <= #1 32'b0;
	else
		tpdw <= #1 tp_in;

endmodule
